#ifndef MRK_PARSER_INTERNAL
#define MRK_PARSER_INTERNAL

#include "mrk/lexer.h"
#include "mrk/parser.h"

#define MRK_PARSER_LOOKAHEAD_BUF_SIZE 4
#define MRK_PARSE_ERR(p, t, m)                                                 \
  p->error.token = t;                                                          \
  p->error.msg = m;                                                            \
  return mrk_err_invalid

struct mrk_parser {
  mrk_lexer *lexer;
  mrk_token next_token;
  // Indentation of the current line
  size_t indent;
  struct {
    mrk_token token;
    const char *msg;
  } error;
};

/**
 * Returns true if all the data from the lexer has been processed.
 *
 * Because the parser might buffer parts of the results of the lexer, this is
 * not equivalent to calling lexer done.
 */
bool mrk_parser_done(mrk_parser *parser);

/**
 * Return the next token that would be consumed by a call to advance.
 */
mrk_token mrk_parser_peek(mrk_parser *parser);

/**
 * Advance the processing stream by one token.
 */
void mrk_parser_advance(mrk_parser *parser);

/**
 * Advance the parser only if the newly consumed token would be the one
 * specified as the argument. Otherwise, throw an error.
 */
mrk_err mrk_parser_eat(mrk_token *out, mrk_parser *parser, mrk_token_type type);

/**
 * Consume all following ident tokens and store how many were seen
 */
void mrk_parser_indent(mrk_parser *parser);

/**
 * Parse all elements on the parser's current indentation level. If the level is
 * 0, this will parse the entire file.
 */
mrk_err mrk_parser_parse_indent_block(mrk_parser *parser, mrk_ast_node *parent);

/**
 * Parse an entire header block.
 */
mrk_err mrk_parser_parse_header(mrk_ast_node *out, mrk_parser *parser);

/**
 * Parse a paragraph. Note that a paragraph does not necessarily end with a
 * blank line, but rather when it either encounters a blank line, end of file,
 * or some other token that indicates a new block should start (e.g. start of a
 * list).
 */
mrk_err mrk_parser_parse_paragraph(mrk_ast_node *out, mrk_parser *parser);

/**
 * Parse a text token, as well as any following text tokens, delimited by a
 * single newline which will be converted to a space.
 */
mrk_err mrk_parser_parse_text(mrk_ast_node *out, mrk_parser *parser);

/**
 * Parse any tokens that have the same meaning no matter the context and append
 * them to the parent. This includes text tokens and links. Only tokens within
 * the same indentation level are matched.
 */
mrk_err mrk_parser_parse_common(mrk_parser *parser, mrk_ast_node *parent);

/**
 * Parse a link construct
 */
mrk_err mrk_parser_parse_link(mrk_ast_node *out, mrk_parser *parser);

/**
 * Parse one of the three list types.
 */
mrk_err mrk_parser_parse_list(mrk_ast_node *out, mrk_parser *parser);

mrk_err mrk_parser_parse_list_item(mrk_ast_node *out, mrk_parser *parser);

#endif
